package main

import (
	"fmt"
	"io/ioutil"
	"os/exec"
	"strings"
	"sync/atomic"
	"syscall"

	"gopkg.in/yaml.v2"
)

type Config struct {
	Settings ConfigSettings `yaml:"settings"`
	Hooks    []Hook         `yaml:"hooks"`
}

type ConfigSettings struct {
	BindAddress string `yaml:"bind"` // Bind address for the webserver
}

type Hook struct {
	Name           string `yaml:"name"`        // name of the hook
	Route          string `yaml:"route"`       // http route
	Command        string `yaml:"command"`     // Actual command to execute
	Background     bool   `yaml:"background"`  // Run in background
	Concurrency    int    `yaml:"concurrency"` // Number of allowed concurrent runs
	concurrentRuns int32  // Number of current concurrent runs
	UID            int    `yaml:"uid"` // UID to use when running the command
	GID            int    `yaml:"gid"` // GID to use when running the command
}

func (cf *Config) SetDefaults() {
	cf.Settings.BindAddress = ":2088"
}

// Check performs sanity checks on the config
func (cf *Config) Check() error {
	if cf.Settings.BindAddress == "" {
		return fmt.Errorf("no bind address configured")
	}
	for _, hook := range cf.Hooks {
		if hook.Name == "" {
			return fmt.Errorf("hook without name")
		}
		if hook.Route == "" {
			return fmt.Errorf("hook %s with no route", hook.Name)
		}
		if hook.Command == "" {
			return fmt.Errorf("hook %s with no command", hook.Name)
		}
		if hook.Concurrency < 1 {
			hook.Concurrency = 1
		}
	}

	return nil
}

func (cf *Config) LoadYAML(filename string) error {
	content, err := ioutil.ReadFile(filename)
	if err != nil {
		return err
	}
	if err := yaml.Unmarshal(content, cf); err != nil {
		return err
	}
	return cf.Check()
}

// Tries to lock a spot. Returns false, if the max. number of concurrent runs has been reached
func (hook *Hook) TryLock() bool {
	res := int(atomic.AddInt32(&hook.concurrentRuns, 1))
	if res > hook.Concurrency {
		atomic.AddInt32(&hook.concurrentRuns, -1)
		return false
	}
	return true
}

func (hook *Hook) Unlock() {
	atomic.AddInt32(&hook.concurrentRuns, -1)
}

// Run executes the given command and return it's return code. It also respects the given concurrency number and will block until resources are free
func (hook *Hook) Run() error {
	split := strings.Split(hook.Command, " ")
	args := make([]string, 0)
	if len(split) > 1 {
		args = split[1:]
	}
	cmd := exec.Command(split[0], args...)
	cmd.SysProcAttr = &syscall.SysProcAttr{}
	cmd.SysProcAttr.Credential = &syscall.Credential{}
	if hook.UID > 0 {
		cmd.SysProcAttr.Credential.Uid = uint32(hook.UID)
	}
	if hook.GID > 0 {
		cmd.SysProcAttr.Credential.Gid = uint32(hook.GID)
	}
	ret := cmd.Run()
	return ret
}
