/*
 * weblug main program
 */
package main

import (
	"fmt"
	"log"
	"net/http"
	"os"
)

var cf Config
var hooks []Hook

type Handler func(http.ResponseWriter, *http.Request)

func usage() {
	fmt.Println("weblug is a simple webhook receiver")
	fmt.Printf("Usage: %s [OPTIONS] YAML1[,YAML2...]\n\n", os.Args[0])
	fmt.Println("OPTIONS")
	fmt.Println("  -h, --help                Print this help message")
	fmt.Println()
	fmt.Println("The program loads the given yaml files for webhook definitions")
}

func main() {
	cf.SetDefaults()
	if len(os.Args) < 2 {
		usage()
	}
	for _, arg := range os.Args[1:] {
		if arg == "" {
			continue
		} else if arg == "-h" || arg == "--help" {
			usage()
			os.Exit(0)
		} else {
			if err := cf.LoadYAML((arg)); err != nil {
				fmt.Fprintf(os.Stderr, "yaml error: %s\n", err)
				os.Exit(1)
			}
		}
	}

	if len(cf.Hooks) == 0 {
		fmt.Fprintf(os.Stderr, "error: no webhooks defined\n")
		os.Exit(2)
	}

	// Create default handlers
	http.HandleFunc("/health", createHealthHandler())
	http.HandleFunc("/health.json", createHealthHandler())
	http.HandleFunc("/index", createDefaultHandler())
	http.HandleFunc("/index.htm", createDefaultHandler())
	http.HandleFunc("/index.html", createDefaultHandler())
	http.HandleFunc("/robots.txt", createRobotsHandler())

	// Register hooks
	for i, hook := range cf.Hooks {
		if hook.Route == "" {
			fmt.Fprintf(os.Stderr, "Invalid hook %s: No route defined\n", hook.Name)
		}
		if hook.Concurrency < 1 {
			hook.Concurrency = 1
		}
		log.Printf("Webhook %d: '%s' [%s] \"%s\"\n", i, hook.Name, hook.Route, hook.Command)
		http.HandleFunc(hook.Route, createHandler(hook))
	}

	// TODO: Await termination signal

	log.Printf("Launching webserver on %s", cf.Settings.BindAddress)
	log.Fatal(http.ListenAndServe(cf.Settings.BindAddress, nil))
}

// create a http handler function from the given hook
func createHandler(hook Hook) Handler {
	return func(w http.ResponseWriter, r *http.Request) {
		log.Printf("GET %s %s", r.RemoteAddr, hook.Name)

		// Check for available slots
		if hook.TryLock() == false {
			log.Printf("ERR: \"%s\" max concurrency reached", hook.Name)
			// 503 - Service Unavailable
			w.Header().Add("Retry-After", "120") // Suggest to retry after 2 minutes
			w.WriteHeader(503)
			fmt.Fprintf(w, "{\"status\":\"fail\",\"reason\":\"max concurrency reached\"}")
			return
		}

		if hook.Background { // Execute command in background
			w.WriteHeader(200)
			fmt.Fprintf(w, "{\"status\":\"ok\"}")
			go func() {
				defer hook.Unlock()
				if err := hook.Run(); err != nil {
					log.Printf("Hook \"%s\" failed: %s", hook.Name, err)
				} else {
					log.Printf("Hook \"%s\" completed", hook.Name)
				}
			}()
		} else {
			defer hook.Unlock()
			if err := hook.Run(); err != nil {
				log.Printf("ERR: \"%s\" exec failure: %s", hook.Name, err)
				w.WriteHeader(500)
				fmt.Fprintf(w, "{\"status\":\"fail\",\"reason\":\"program error\"}")
			} else {
				w.WriteHeader(200)
				fmt.Fprintf(w, "{\"status\":\"ok\"}")
			}
		}
	}
}

func createHealthHandler() Handler {
	return func(w http.ResponseWriter, r *http.Request) {
		w.WriteHeader(200)
		fmt.Fprintf(w, "{\"status\":\"ok\"}")
	}
}

func createDefaultHandler() Handler {
	return func(w http.ResponseWriter, r *http.Request) {
		w.WriteHeader(200)
		fmt.Fprintf(w, "weblug")
	}
}

func createRobotsHandler() Handler {
	return func(w http.ResponseWriter, r *http.Request) {
		w.WriteHeader(200)
		fmt.Fprintf(w, "User-agent: *\nDisallow: /")
	}
}
